/*
 * Created on May 14, 2007
 */
package com.x8ing.lsm4j.state;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Stores the structure of a graph. It must be possible to have several tansitions between two nodes.
 * <p>
 * We need two ways to fast get a transition.
 * <p>
 * 1. get all transitions starting from a node A.<br>
 * 2. get all transition between node A and node B <br>
 *
 * We make a memory/speed tradeoff and use two hashmaps: <br>
 * 1. internode hashmap <br>
 * 2. outgoing node hashmap
 *
 *
 * @author Patrick Heusser
 */
public class StaticGraph {

    private static final String KEY_SEP = ",";

    /**
     * key: String using method getInternodeKey() <br>
     * value: TransitionList
     */
    private Map internodes = new HashMap();

    /**
     * key: String using method getOutgoingKey() <br>
     * value: TransitionList
     */
    private Map outgoings = new HashMap();

    /**
     * A list of transitions.
     *
     * @author Patrick Heusser
     */
    protected static class TransitionList {

        /**
         * type: StaticTransition
         */
        private List transitions = new ArrayList();

        protected void addTransition(StaticTransition transition) {
            transitions.add(transition);
        }

        protected List getTransitionsList() {
            return transitions;
        }

    }

    /**
     * TODO implement stateiterator and transition iterator!!!
     *
     * Iterates over all states in the graph.
     *
     * @author Patrick Heusser
     */
    protected static class StateIterator implements Iterator {

        private Map outgoings = null;

        private Iterator itOutGoings = null;

        private StaticState nextReturn = null;

        protected StateIterator(Map outgoings) {
            this.outgoings = outgoings;
        }

        public boolean hasNext() {
            return false;
        }

        /**
         * Type: {@link StaticState}
         */
        public Object next() {

            return null;
        }

        private void prepareNextReturn() {

        }

        /**
         * Not supported. Iterator is read only!
         *
         * @throws UnsupportedOperationException
         */
        public void remove() {

            // TODO may be implement once...
            throw new UnsupportedOperationException("graph is read only. remove is not allowed.");

        }
    }

    /**
     * @return null if not found
     */
    protected StaticState getStateWithID(int stateUID) {
        List list = getTransitionListForState(stateUID);

        if(list==null || list.isEmpty()) {
            return null;
        }

        StaticTransition trans = (StaticTransition) list.get(0);

        if (trans == null) {
            return null;
        }

        return trans.getCurrentState();
    }

    /**
     *
     */
    public void addTransition(StaticTransition transition) {

        // add it to internodes
        {
            String keyInternode = getInternodeKey(transition);
            TransitionList internodeTransitionList = (TransitionList) internodes.get(keyInternode);
            if (internodeTransitionList == null) {
                // lazy add one entry
                internodeTransitionList = new TransitionList();
            }
            internodeTransitionList.addTransition(transition);
            internodes.put(keyInternode, internodeTransitionList);
        }

        // add it to outgoing nodes
        {
            String keyOutgoing = getOutgoingKey(transition);
            TransitionList outgoingTransitionList = (TransitionList) outgoings.get(keyOutgoing);
            if (outgoingTransitionList == null) {
                // lazy add one entry
                outgoingTransitionList = new TransitionList();
            }
            outgoingTransitionList.addTransition(transition);
            outgoings.put(keyOutgoing, outgoingTransitionList);

        }

    }

    /**
     * Get a list of transitions that leave a certain state.
     *
     * If nothing is found, an empty list is returned.
     *
     * @param stateUID
     * @return List with elements of type: {@link StaticTransition}
     */
    public List getTransitionListForState(int stateUID) {

        String key = getOutgoingKey(stateUID);

        TransitionList transitionList = (TransitionList) outgoings.get(key);

        // return empty list
        if (transitionList == null) {
            return new ArrayList();
        }

        return transitionList.getTransitionsList();

    }

    /**
     * Get a list of transitions that are between two states.
     *
     * If nothing is found, an empty list is returned.
     *
     * @param currentStateUID
     * @param nextStateUID
     * @return List with elements of type: {@link StaticTransition}
     */
    public List getTransitionsBetweenStates(int currentStateUID, int nextStateUID) {

        String key = getInternodeKey(currentStateUID, nextStateUID);

        TransitionList transitionList = (TransitionList) internodes.get(key);

        // return empty list
        if (transitionList == null) {
            return new ArrayList();
        }

        return transitionList.getTransitionsList();

    }

    /**
     * @return currentNode.UID + SEP + nextNode.UID
     */
    private static String getInternodeKey(StaticTransition transition) {

        return getInternodeKey(transition.currentState.getUniqueID(), transition.nextState.getUniqueID());
    }

    /**
     * @return currentNode.UID + SEP + nextNode.UID
     */
    private static String getInternodeKey(int currentUID, int nextUID) {

        String key = String.valueOf(currentUID) + KEY_SEP + String.valueOf(nextUID);

        return key;
    }

    /**
     * @return currentNode.UID
     */
    private static String getOutgoingKey(StaticTransition transition) {

        return getOutgoingKey(transition.currentState.getUniqueID());
    }

    /**
     * @return currentNode.UID
     */
    private static String getOutgoingKey(int stateUID) {

        String key = String.valueOf(stateUID);

        return key;
    }

    public String printGraph() {

        StringBuffer ret = new StringBuffer();

        ret.append("\n");
        for (Iterator it1 = outgoings.values().iterator(); it1.hasNext();) {
            TransitionList transitionList = (TransitionList) it1.next();
            for (Iterator it2 = transitionList.getTransitionsList().iterator(); it2.hasNext();) {

                StaticTransition staticTransition = (StaticTransition) it2.next();
                ret.append("\n");
                ret.append(staticTransition);

            }

        }

        return ret.toString();
    }

    /**
     *
     * @return the count of transitions in the graph.
     */
    public int getTransitionsCount() {

        return internodes.size();

    }

    /**
     *
     * @return the count of states in the graph.
     */
    public int getStateCount() {

        return outgoings.size();
    }

    public void addValidTransition(StaticTransition transition) {
        addTransition(transition);
    }

    public void addValidTransition(StaticState currentState, StaticState nextState) {
        StaticTransition transition = new StaticTransition(currentState, nextState);
        addValidTransition(transition);
    }

    public boolean isValidTransition(StaticTransition transition) {
        int uid1 = transition.currentState.getUniqueID();
        int uid2 = transition.nextState.getUniqueID();
        List list = getTransitionsBetweenStates(uid1, uid2);
        return list.isEmpty();
    }

    public boolean isValidTransition(StaticState currentState, StaticState nextState) {
        int uid1 = currentState.getUniqueID();
        int uid2 = nextState.getUniqueID();
        List list = getTransitionsBetweenStates(uid1, uid2);
        return !list.isEmpty();

    }

    public String toString() {
        StringBuffer ret = new StringBuffer();
        ret.append(printGraph());
        ret.append("\n stateCount=");
        ret.append(getStateCount());
        ret.append("\n transitionsCount=");
        ret.append(getTransitionsCount());
        ret.append("\n");

        return ret.toString();
    }
}
