/*
 * Created on Aug 29, 2007
 */
package com.x8ing.mc.bp;

import java.text.NumberFormat;
import java.util.Calendar;
import java.util.List;

import com.x8ing.lsm4j.Action;
import com.x8ing.lsm4j.Condition;
import com.x8ing.lsm4j.StateContext;
import com.x8ing.lsm4j.state.ProcessableState;

/**
 * Provides some functionality which is useful in context of business processes.
 * <p>
 * Controls the simulation time.
 * <p>
 * 
 * @author Patrick Heusser
 * 
 */
public abstract class AbstractBusinessAction implements Action {

	private NumberFormat numberFormat = Constants.createCurrencyFormater();

	/**
	 * subclasses might use this.
	 */
	protected BusinessContext businessContext = null;

	protected ProcessableState currentState = null;

	private boolean lazyInitDone = false;

	/**
	 * @see com.x8ing.lsm4j.Action#execute(com.x8ing.lsm4j.state.ProcessableState, com.x8ing.lsm4j.StateContext,
	 *      com.x8ing.lsm4j.Condition, java.util.List)
	 */
	final public void execute(ProcessableState currentState, StateContext stateContext, Condition previousCondition,
			List lastVisitedStatesHistory) {

		this.businessContext = (BusinessContext) stateContext;
		this.currentState = currentState;

		// create an init hook, after we have the business context!
		if (!lazyInitDone) {
			lazyInit();
			lazyInitDone = true;
		}

		execute(currentState, businessContext, previousCondition, lastVisitedStatesHistory);

		moveSimulationTimeForwardBy(24);

	}

	/**
	 * hook just before the execute will be executed. will just be called once in the lifetime of an action.
	 */
	protected abstract void lazyInit();

	/**
	 * move the simulation time forward for a certain time. this depends on the action that must be taken within the
	 * concrete implementing action. the value reflects the real time that would be needed in reality.
	 * 
	 * @param hours
	 */
	private void moveSimulationTimeForwardBy(int hours) {

		businessContext.getCurrentDate().add(Calendar.HOUR, hours);

	}

	/**
	 * abstract to force implementors to implement that method.
	 * 
	 * @param currentState
	 * @param businessContext
	 * @param previousCondition
	 * @param lastVisitedStatesHistory
	 */
	public abstract void execute(ProcessableState currentState, BusinessContext businessContext, Condition previousCondition,
			List lastVisitedStatesHistory);

	/**
	 * convenience method.
	 * 
	 * @param text
	 */
	protected void addLogBookEntry(String text) {

		Calendar currentDate = businessContext.getCurrentDate();

		text += ("  [" + currentState.getDescription() + "]");

		businessContext.getLogBook().addLogEntry(currentDate, text);

	}

	protected void addBalanceSheetTransaction(String text, double amount, BalanceAccount balanceAccount) {

		Calendar currentDate = businessContext.getCurrentDate();
		businessContext.getBalanceSheet().addTransaction(text, amount, currentDate, balanceAccount);
		businessContext.getLogBook().addLogEntry(currentDate, "balanceSheet: " + text + " amount=" + numberFormat.format(amount));

	}

}
