/*
 * Created on Aug 29, 2007
 *
 */
package com.x8ing.mc.bp;

import java.io.Serializable;

import com.x8ing.lsm4j.state.MaximumIterationsReachedException;
import com.x8ing.lsm4j.state.NoMatchingTransitionConditionFoundException;
import com.x8ing.lsm4j.state.ProcessableGraph;
import com.x8ing.lsm4j.state.ProcessableState;
import com.x8ing.lsm4j.std.TrueCondition;
import com.x8ing.mc.Configuration;
import com.x8ing.mc.bp.develop.AddBugToFixListAction;
import com.x8ing.mc.bp.develop.AnalyzeCostLossAction;
import com.x8ing.mc.bp.develop.BugFixedCondition;
import com.x8ing.mc.bp.develop.CostLossCheckCondition;
import com.x8ing.mc.bp.develop.DeployAndReleaseAction;
import com.x8ing.mc.bp.develop.DeployAndReleaseCondition;
import com.x8ing.mc.bp.develop.DocumentAndFinishProcessAction;
import com.x8ing.mc.bp.develop.EndProcessAction;
import com.x8ing.mc.bp.develop.FixAllKnownBugsAction;
import com.x8ing.mc.bp.develop.TestingAction;
import com.x8ing.mc.bp.develop.TestingPassedCondition;
import com.x8ing.mc.bp.operation.OperationDownAction;
import com.x8ing.mc.bp.operation.OperationRunningDependentCondition;
import com.x8ing.mc.bp.operation.OperationRunsAction;

/**
 * @author Patrick Heusser
 */
public class BusinessProcesses implements Serializable {

	private static final long serialVersionUID = 4384932097601999397L;

	private Configuration configuration = null;

	private ProcessableGraph bpOperation = null;

	private ProcessableGraph bpChange = null;

	static final int bpStartID = 1;

	public BusinessProcesses(Configuration configuration) {
		this.configuration = configuration;
		init();
	}

	private void init() {
		setUpProcessGraph();
	}

	private void setUpProcessGraph() {

		int uidCounter = 100;

		// scope: set up operation process
		{

			bpOperation = new ProcessableGraph();

			// //////////////////////////////////////////////////////////////////////////////
			// create actions
			OperationRunsAction operationRunsAction = new OperationRunsAction();
			OperationDownAction operationDownAction = new OperationDownAction();

			// //////////////////////////////////////////////////////////////////////////////
			// create states
			ProcessableState operationRuns = new ProcessableState(bpStartID, "operationRuns", operationRunsAction, false);
			ProcessableState operationDown = new ProcessableState(uidCounter++, "operationRuns", operationDownAction, false);

			// //////////////////////////////////////////////////////////////////////////////
			// create transitions
			bpOperation.addValidTransition(operationRuns, operationRuns, new OperationRunningDependentCondition(true));
			bpOperation.addValidTransition(operationRuns, operationDown, new OperationRunningDependentCondition(false));
			bpOperation.addValidTransition(operationDown, operationDown, new OperationRunningDependentCondition(false));
			bpOperation.addValidTransition(operationDown, operationRuns, new OperationRunningDependentCondition(true));
		}

		// scope: set up change process
		{
			bpChange = new ProcessableGraph();

			// //////////////////////////////////////////////////////////////////////////////
			// create actions
			AnalyzeCostLossAction analyzeCostLossAction = new AnalyzeCostLossAction();
			AddBugToFixListAction addBugToFixListAction = new AddBugToFixListAction();
			FixAllKnownBugsAction fixAllKnownBugsAction = new FixAllKnownBugsAction();
			DeployAndReleaseAction deployAndReleaseAction = new DeployAndReleaseAction();
			DocumentAndFinishProcessAction documentAndFinishProcessAction = new DocumentAndFinishProcessAction();
			TestingAction testingAction = new TestingAction();
			EndProcessAction endProcessAction = new EndProcessAction();

			// //////////////////////////////////////////////////////////////////////////////
			// create states
			ProcessableState analyzeCostLoss = new ProcessableState(bpStartID, "analyzeCostLoss", analyzeCostLossAction, false);
			ProcessableState addBugToFixList = new ProcessableState(uidCounter++, "addBugToFixList", addBugToFixListAction, false);
			ProcessableState fixAllKnownBugs = new ProcessableState(uidCounter++, "fixAllKnownBugs", fixAllKnownBugsAction, false);
			ProcessableState testing = new ProcessableState(uidCounter++, "testing", testingAction, false);
			ProcessableState deployAndRelease = new ProcessableState(uidCounter++, "deployAndRelease", deployAndReleaseAction,
					false);
			ProcessableState documentAndFinishProcess = new ProcessableState(uidCounter++, "documentAndFinishProcess",
					documentAndFinishProcessAction, false);
			ProcessableState endState = new ProcessableState(uidCounter++, "end", endProcessAction, false);

			// //////////////////////////////////////////////////////////////////////////////
			// create transitions
			bpChange.addValidTransition(analyzeCostLoss, addBugToFixList, new CostLossCheckCondition(true, configuration
					.getProcessCostLossTriggerForBugFixing()));
			bpChange.addValidTransition(analyzeCostLoss, fixAllKnownBugs, new CostLossCheckCondition(false, configuration
					.getProcessCostLossTriggerForBugFixing()));

			bpChange.addValidTransition(fixAllKnownBugs, testing, new BugFixedCondition(false));
			bpChange.addValidTransition(fixAllKnownBugs, fixAllKnownBugs, new BugFixedCondition(true));

			bpChange.addValidTransition(testing, deployAndRelease, new TestingPassedCondition(false));
			bpChange.addValidTransition(testing, fixAllKnownBugs, new TestingPassedCondition(true));

			bpChange.addValidTransition(deployAndRelease, deployAndRelease, new DeployAndReleaseCondition(true));
			bpChange.addValidTransition(deployAndRelease, documentAndFinishProcess, new DeployAndReleaseCondition(false));

			bpChange.addValidTransition(addBugToFixList, documentAndFinishProcess, new TrueCondition());

			bpChange.addValidTransition(documentAndFinishProcess, endState, new TrueCondition()); // end

		}

	}

	public BusinessContext run() {

		BusinessContext businessContext = new BusinessContext(configuration);

		try {

			// init both processes
			bpOperation.run(bpStartID, businessContext, 1);

			for (int i = 0; i < configuration.getNumberOfSimulationDays(); i++) {

				// let the operation run for one step
				bpOperation.runContinue(1);

				// check for new bugs
				int queuedBugs = businessContext.getQueuedBugs().size();
				boolean isDevelopProcessBusy = businessContext.isDevelopProcessBusy();

				if (isDevelopProcessBusy) {

					// let developers do their work
					businessContext.getLogBook().addLogEntry(businessContext.getCurrentDate(),
							"developers are still busy, let them work.");

					bpChange.runContinue(1);

				} else {

					// develop department has free capacity, check for any queued bugs
					if (queuedBugs > 0) {

						businessContext.getLogBook().addLogEntry(
								businessContext.getCurrentDate(),
								"we have " + queuedBugs
										+ " bugs in queue. develop department has free capacity, we give analyzeCostLoss order.");

						// transfer bugs to develop process
						businessContext.transferQueuedBugsToDevelopBugList();

						// start parallel change-process
						bpChange.run(bpStartID, businessContext, 1);
					}

				}

			}

		} catch (NoMatchingTransitionConditionFoundException e) {
			e.printStackTrace();
		} catch (MaximumIterationsReachedException e) {
			e.printStackTrace();
		}

		return businessContext;

	}

}
