/*
 * Created on May 14, 2007
 */
package com.x8ing.lsm4j.processable;

import java.util.List;

import com.x8ing.lsm4j.Condition;
import com.x8ing.lsm4j.Action;
import com.x8ing.lsm4j.StateContext;
import com.x8ing.lsm4j.state.MaximumIterationsReachedException;
import com.x8ing.lsm4j.state.NoMatchingTransitionConditionFoundException;
import com.x8ing.lsm4j.state.ProcessableGraph;
import com.x8ing.lsm4j.state.ProcessableState;
import com.x8ing.lsm4j.std.TracingGraphListener;

import junit.framework.TestCase;

/**
 *
 * @author Patrick Heusser
 */
public class ProcessableGraphTest2 extends TestCase {

	/**
	 * Temperature control loop.
	 */
	public void test1() {

		ProcessableGraph graph = new ProcessableGraph();
		graph.registerGraphListener(new TracingGraphListener(System.out));

		TemperatureControlLoop heat = new TemperatureControlLoop(1);
		TemperatureControlLoop heatStrong = new TemperatureControlLoop(3);
		TemperatureControlLoop cool = new TemperatureControlLoop(-1);
		TemperatureControlLoop doNothing = new TemperatureControlLoop(0);

		ProcessableState stateA = new ProcessableState(1, "stateA", doNothing, false); // start
		ProcessableState stateB = new ProcessableState(2, "stateB", heatStrong, false);
		ProcessableState stateC = new ProcessableState(3, "stateC", heat, false);
		ProcessableState stateD = new ProcessableState(4, "stateD", cool, false);
		ProcessableState stateE = new ProcessableState(5, "stateE", doNothing, true); // end

		// add transitions: leaving from A
		graph.addValidTransition(stateA, stateB, new TemperatureCondition("freezing"));
		graph.addValidTransition(stateA, stateC, new TemperatureCondition("cold"));
		graph.addValidTransition(stateA, stateD, new TemperatureCondition("warm"));
		graph.addValidTransition(stateA, stateE, new TemperatureCondition("optimum"));

		// add transitions: leaving from B
		graph.addValidTransition(stateB, stateA, new TemperatureCondition("any"));

		// add transitions: leaving from C
		graph.addValidTransition(stateC, stateA, new TemperatureCondition("any"));

		// add transitions: leaving from D
		graph.addValidTransition(stateD, stateA, new TemperatureCondition("any"));

		System.out.println(graph);

		{
			TemperatureContext tempContext = new TemperatureContext();

			// start very cold
			tempContext.currentTemperature = -10;

			try {

				ProcessableState lastState = graph.run(1, tempContext);

				// assertEquals(5, tempContext.currentTemperature);

				assertEquals(lastState.getUniqueID(), 5);

			} catch (NoMatchingTransitionConditionFoundException e) {
				fail();
			} catch (MaximumIterationsReachedException e) {
				fail();
			}
		}

		{
			TemperatureContext tempContext = new TemperatureContext();

			// start to warm
			tempContext.currentTemperature = 25;

			try {

				ProcessableState lastState = graph.run(1, tempContext);

				// assertEquals(5, tempContext.currentTemperature);

				assertEquals(lastState.getUniqueID(), 5);

			} catch (NoMatchingTransitionConditionFoundException e) {
				fail();
			} catch (MaximumIterationsReachedException e) {
				fail();
			}
		}

	}

	private static void trace(String msg) {
		System.out.println("\t " + msg);

	}

	private static class TemperatureControlLoop implements Action {

		private int temperatureChange = 1;

		public TemperatureControlLoop(int temperatureChange) {
			this.temperatureChange = temperatureChange;
		}

		private TemperatureContext temperatureContext = null;

		private ProcessableState currentState = null;

		private TemperatureCondition temperatureCondition = null;

		public void execute(ProcessableState currentState, StateContext stateContext, Condition previousCriteria,
				List lastVisitedStatesHistory) {
			this.temperatureContext = (TemperatureContext) stateContext;
			this.currentState = currentState;
			this.temperatureCondition = (TemperatureCondition) previousCriteria;

			// adjust temperature
			temperatureContext.currentTemperature += temperatureChange;

		}

	}

	private static class TemperatureContext implements StateContext {

		int currentTemperature = -1;

	}

	private static class TemperatureCondition implements Condition {

		private static final String WARM = "warm";

		private static final String COLD = "cold";

		private static final String FREEZING = "freezing";

		private static final String OPTIMUM = "optimum";

		// matches all conditions
		private static final String ANY = "any";

		/**
		 * can be WARM, COLD, FREEZING, OPTIMUM or ANY
		 */
		private String verbalDescription = null;

		/**
		 * @param simpleCriteria
		 */
		public TemperatureCondition(String verbalDescription) {
			super();
			this.verbalDescription = verbalDescription;
		}

		public String traceInfo() {
			return "verbalDescription=" + verbalDescription;
		}

		/**
		 * @see com.x8ing.lsm4j.Condition#conditionTrue(com.x8ing.lsm4j.StateContext)
		 */
		public boolean conditionTrue(StateContext currentContext) {

			TemperatureContext tempContext = (TemperatureContext) currentContext;

			// transform the temperature in the verbal form
			int temperature = tempContext.currentTemperature;

			String verbalTemp = null;

			// check for asterix (any) condition
			if (verbalDescription.equalsIgnoreCase(ANY)) {
				return true;
			}

			// dispatch temperature
			if (temperature < 0) {
				verbalTemp = FREEZING;
			} else if (temperature >= 0 && temperature < 18) {
				verbalTemp = COLD;
			} else if (temperature >= 18 && temperature < 22) {
				verbalTemp = OPTIMUM;
			} else if (temperature >= 22) {
				verbalTemp = WARM;
			} else {
				// not found
				throw new RuntimeException("could not figure out verbal description for temperature=" + temperature);
			}

			boolean match = verbalTemp.equalsIgnoreCase(verbalDescription);

			if (match) {
				trace("condition: temperature=" + temperature + " is " + verbalTemp + " conditionMatch=" + match);
			}

			return match;
		}

	}

}
