/*
 * Created on May 14, 2007
 */
package com.x8ing.lsm4j.state;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.x8ing.lsm4j.Action;
import com.x8ing.lsm4j.Condition;
import com.x8ing.lsm4j.GraphListener;
import com.x8ing.lsm4j.StateContext;

/**
 * This class is the entry point for building a state machine. It keeps track of the graph layout and acts as controller
 * for running it.
 * <p>
 * How to use this class is described in the tutorial <a href="./tutorial.html"> link </a>.
 *
 * @author Patrick Heusser
 */
public class ProcessableGraph extends StaticGraph {

	/**
	 * A list with all listeners.
	 * <p>
	 * type: {@link GraphListener}
	 */
	private List graphListeners = new ArrayList();

	/**
	 * It might be useful to specify a maximum loop amount. If more then the given numbers of transitions were executed,
	 * the processing stops with a MaximumIterationsReachedException.
	 */
	private long maximumLoops = -1;

	/**
	 * defines the size of the last visited states list.
	 */
	private int lastVisitedStatesHistorySize = 5;

	/**
	 * A list with the last visited states. The history where we came from. Index 0 means the current state.
	 * <p>
	 * element type: {@link ProcessableState}
	 */
	private List lastVisitedStatesHistoryList = new LinkedList();

	/**
	 * The last state that was processed.
	 */
	private ProcessableState currentState = null;

	/**
	 * The current context.
	 */
	private StateContext currentStateContext = null;

	public void addValidTransition(ProcessableState currentState, ProcessableState nextState, Condition condition,
			Action transitionAction) {

		ProcessableTransition processableTransition = new ProcessableTransition(currentState, nextState, condition,
				transitionAction);
		super.addValidTransition(processableTransition);

	}

	public void addValidTransition(ProcessableState currentState, ProcessableState nextState, Condition condition) {

		addValidTransition(currentState, nextState, condition, null);

	}

	/**
	 * Continues the processing of the previously interrupted graph with a limited amount of transitions.
	 *
	 * @param numberOfSteps
	 *            the number of transitions to be executed.
	 * @return the last State or null if entry point was not found.
	 * @throws NoMatchingTransitionConditionFoundException
	 * @throws MaximumIterationsReachedException
	 */
	public ProcessableState runContinue(int numberOfSteps) throws NoMatchingTransitionConditionFoundException,
			MaximumIterationsReachedException {

		if (currentState == null) {
			String diagnostic = printLastVisitedHistory(5);
			handleErrorStateNotFound(0, "state to continue operation not found. diagnosticHistory=" + diagnostic);

		}

		return runImpl(currentState.getUniqueID(), numberOfSteps);
	}

	/**
	 * Continues the processing of the previously interrupted graph.
	 *
	 * @return the last State or null if entry point was not found.
	 * @throws NoMatchingTransitionConditionFoundException
	 * @throws MaximumIterationsReachedException
	 */
	public ProcessableState runContinue() throws NoMatchingTransitionConditionFoundException, MaximumIterationsReachedException {
		return runContinue(-1);
	}

	/**
	 * Starts the processing of the graph. The processing ends if an end point is found.
	 *
	 * Pleaes not that the startStateContext will be changed since it's passed by reference!
	 *
	 * @param startStateID
	 * @param startStateContext
	 * @throws NoMatchingTransitionConditionFoundException
	 * @throws MaximumIterationsReachedException
	 * @return the last State or null if entry point was not found.
	 */
	public ProcessableState run(int startStateID, StateContext startStateContext)
			throws NoMatchingTransitionConditionFoundException, MaximumIterationsReachedException {

		return run(startStateID, startStateContext, -1);

	}

	/**
	 * Run the graph for a defined number of transitions.
	 *
	 * Pleae not that the startStateContext will be changed since it's passed by reference!
	 *
	 * @param startStateID
	 * @param startStateContext
	 * @param numberOfSteps
	 *            the number of transitions to be executed.
	 * @throws NoMatchingTransitionConditionFoundException
	 * @throws MaximumIterationsReachedException
	 * @return the last State or null if entry point was not found.
	 */
	public ProcessableState run(int startStateID, StateContext startStateContext, int numberOfSteps)
			throws NoMatchingTransitionConditionFoundException, MaximumIterationsReachedException {

		currentStateContext = startStateContext;
		return runImpl(startStateID, numberOfSteps);

	}

	private ProcessableState runImpl(int startStateID, int numberOfSteps) throws MaximumIterationsReachedException,
			NoMatchingTransitionConditionFoundException {

		currentState = (ProcessableState) getStateWithID(startStateID);
		if (currentState == null) {
			return null;
		}

		Condition previousCondition = null;
		Condition nextCondition = null;

		long loop = 0;

		while (true) {
			loop++;

			// notify listeners: start processing
			for (Iterator it = graphListeners.iterator(); it.hasNext();) {
				GraphListener listener = (GraphListener) it.next();
				listener.startProcessingState(currentState, previousCondition, loop, currentStateContext);
			}

			// check for exceeding of max loops
			if (maximumLoops != -1 && loop >= maximumLoops) {
				MaximumIterationsReachedException ex = new MaximumIterationsReachedException();
				ex.setMaximumIterationsSpecified(maximumLoops);
				throw ex;
			}

			if (currentState == null) {
				handleErrorStateNotFound(startStateID, "State not found");
			}

			// prepare history
			addStateToLastVisitedHistory(currentState);

			// invoke action
			currentState.getAction().execute(currentState, currentStateContext, previousCondition, lastVisitedStatesHistoryList);

			// check end state
			if (currentState.isEndState()) {

				// notify listeners: end
				for (Iterator it = graphListeners.iterator(); it.hasNext();) {
					GraphListener listener = (GraphListener) it.next();
					listener.foundEndState(currentState, loop, currentStateContext);
				}

				// EXIT POINT
				return currentState;

			}

			// check for a condition which is true
			ProcessableTransition nextProcessableTransition = null;

			for (Iterator it = getTransitionListForState(currentState.getUniqueID()).iterator(); it.hasNext();) {

				ProcessableTransition processableTransition = (ProcessableTransition) it.next();

				nextCondition = processableTransition.getCondition();

				if (nextCondition.conditionTrue(currentStateContext)) {
					// found a transition to go
					nextProcessableTransition = processableTransition;
					break;
				}
			}

			if (nextProcessableTransition == null) {
				NoMatchingTransitionConditionFoundException ex = new NoMatchingTransitionConditionFoundException();

				ex.setCurrentProcessableTransition(null);
				ex.setState(currentState);

				throw ex;
			}

			ProcessableState nextState = nextProcessableTransition.getNextProcessableState();

			// notify listeners: changed state
			for (Iterator it = graphListeners.iterator(); it.hasNext();) {
				GraphListener listener = (GraphListener) it.next();
				listener.changedState(currentState, nextState, previousCondition, nextCondition, loop, currentStateContext);
			}

			// last: invoke an optional action on the transition
			Action transitionAction = nextProcessableTransition.getAction();
			if (transitionAction != null) {
				transitionAction.execute(currentState, currentStateContext, previousCondition, lastVisitedStatesHistoryList);
			}

			// assign variables
			currentState = nextState;
			previousCondition = nextCondition;

			// at the very end: check for limit of steps
			if (numberOfSteps != -1 && loop >= numberOfSteps) {
				return currentState;
			}

		}
	}

	private void addStateToLastVisitedHistory(ProcessableState currentState) {

		lastVisitedStatesHistoryList.add(0, currentState);

		// check max size
		if (lastVisitedStatesHistoryList.size() > lastVisitedStatesHistorySize) {
			lastVisitedStatesHistoryList.remove(lastVisitedStatesHistoryList.size() - 1);
		}
	}

	public void registerGraphListener(GraphListener graphListener) {

		graphListeners.add(graphListener);

	}

	/**
	 * @param graphListener
	 *            to be removed.
	 * @return true if remove was success.
	 */
	public boolean unregisterGraphListener(GraphListener graphListener) {

		return graphListeners.remove(graphListener);

	}

	private void handleErrorStateNotFound(int stateID, String text) {
		throw new RuntimeException("State not found with id=" + stateID + " info=" + text);
	}

	public long getMaximumLoops() {
		return maximumLoops;
	}

	public void setMaximumLoops(long maximumLoops) {
		this.maximumLoops = maximumLoops;
	}

	public int getLastVisitedStatesHistorySize() {
		return lastVisitedStatesHistorySize;
	}

	public void setLastVisitedStatesHistorySize(int lastVisitedStatesHistorySize) {
		this.lastVisitedStatesHistorySize = lastVisitedStatesHistorySize;
	}

	/**
	 * A list with all listeners.
	 * <p>
	 * type: {@link GraphListener}
	 */
	public List getGraphListeners() {
		return graphListeners;
	}

	/**
	 * A list with the last visited states. The history where we came from. Index 0 means the current state.
	 * <p>
	 * element type: {@link ProcessableState}
	 */
	public List getLastVisitedStatesHistoryList() {
		return lastVisitedStatesHistoryList;
	}

	private String printLastVisitedHistory(int maxEntries) {

		StringBuffer ret = new StringBuffer();
		ret.append("lastVisitedHistory { ");

		if (lastVisitedStatesHistoryList != null && !lastVisitedStatesHistoryList.isEmpty()) {
			int i = 0;

			for (Iterator it = lastVisitedStatesHistoryList.iterator(); it.hasNext();) {
				i++;
				if (i > maxEntries) {
					break;
				}
				ProcessableState state = (ProcessableState) it.next();
				ret.append("entry #-");
				ret.append(i);
				ret.append(",");
				if (state == null) {
					ret.append("StateIsNull");
				} else {
					ret.append("id=");
					ret.append(state.getUniqueID());
					ret.append(", desc=");
					ret.append(state.getDescription());
				}
				ret.append(";  ");
			}
		}

		ret.append("}");

		return ret.toString();

	}
}
